/*

LRGMP_int_ext.c -- extension to the LRGMP integer functions for the 
                   iRRAM library with backend LRGMP
 
Copyright (C) 2001 Norbert Mueller
 
This file is part of the iRRAM Library.
 
The iRRAM Library is free software; you can redistribute it and/or modify
it under the terms of the GNU Library General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at your
option) any later version.
 
The iRRAM Library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
License for more details.
 
You should have received a copy of the GNU Library General Public License
along with the iRRAM Library; see the file COPYING.LIB.  If not, write to
the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
MA 02111-1307, USA. 
*/

#include "LRGMP_int_ext.h"

#include <strings.h>

#ifndef DEBUG			// debugging information
#define DEBUG 0
#endif
#if DEBUG > 0
  long DEBUG_OP = 1;
  unsigned long op_count=0;
  long minshowop = DEBUG ;
#endif

//*****************************************************************************
// Variables for counting free space
//*****************************************************************************

#define MaxFreeVars 1000
static int_gmp_type gmp_FreeVarsi[MaxFreeVars];
static long gmp_FreeVarCounti=0L;
long int_gmp_var_count=0;


//*****************************************************************************
// INITIALIZING FUNCTIONS
//*****************************************************************************
// initialize GMP_integer, allocating memory
// added: 31.01.2001
// Last change: 31.01.2001
// Arguments: none
// 		  return: GMP_integer
//*****************************************************************************

int_gmp_type int_gmp_init()
{
	int_gmp_type z;
	if (gmp_FreeVarCounti>0) 
		{
		gmp_FreeVarCounti--;
		z = gmp_FreeVarsi[gmp_FreeVarCounti];
		}
	else
		{
		z = (int_gmp_type) malloc(sizeof(__mpz_struct));
		mpz_init(z);
		}
	int_gmp_var_count++;
	return z;
}





//*****************************************************************************
// delete no longer used GMP_integers, including deallocation
// added: 31.01.2001
// Last change: 31.01.2001
// Arguments: 1. GMP_integer
//*****************************************************************************

void int_gmp_free(int_gmp_type z)
{
  if (gmp_FreeVarCounti < MaxFreeVars)
  		{
  		gmp_FreeVarsi[gmp_FreeVarCounti]=z;
  		gmp_FreeVarCounti++;
  		}
  	else
  	{
  	mpz_clear(z);
  	free(z);
  	}
  int_gmp_var_count--;

  //mpz_clear(z);
}


//*****************************************************************************
// Getting size of GMP_integers
// Added: 01.05.2001
// Last change: 01.05.2001
// Arguments: 1. GMP_integer
// 		  return: (standard) integer
//*****************************************************************************

long int_gmp_getsize(int_gmp_type z) //****** Anzahl der Ziffern
{
unsigned long r=0,w,n,m;
if (z->_mp_size==0) return GMP_min;
w=(z->_mp_d)[abs(z->_mp_size)-1];

n=w;
for (m=mp_bits_per_limb>>1; m>0; m=m>>1 ) {
      if (n>= (1<<m) ){
        r+=m; n=n>>m;
      }
    }
if (w > (1<<r) || abs(z->_mp_size) >0) r++;    

return 1+r+8* sizeof (mp_limb_t);
}

//*****************************************************************************
// CONVERSION FUNCTIONS
//*****************************************************************************
// Converting standart integer (int) to MP_integer
// added: 01.05.2001
// Last change: 01.05.2001
// Arguments: 1. signed (standard) integer
//		  2. GMP_integer
//*****************************************************************************

void int_gmp_int2int(long int i, int_gmp_type z)
{
 mpz_set_si(z,i);
 return;
}

//*****************************************************************************
// Converting standard double to MP_integer
// added: 01.05.2001
// Last change: 01.05.2001
// Arguments: 1. (standard) double
//		  2. GMP_integer
//*****************************************************************************

void int_gmp_double2int(double d, int_gmp_type z)
{
 mpz_set_d(z,d);
 return;
}

//*****************************************************************************
// Converting string to MP_integer
// added: 01.05.2001
// Last change: 01.05.2001
// Arguments: 1. string
// 		  2. GMP_integer
// 		  3. Base (i.e. 10 for decimal)
//*****************************************************************************

void int_gmp_string2int(char* s, int_gmp_type z, int b)
{
 mpz_set_str(z,s,b);
 return;
}

//*****************************************************************************
// Converting GMP_real to GMP_integer
// added: 01.05.2001
// Last change: 01.05.2001
// Arguments: 1. GMP_real
// 		  2. GMP_integer
//*****************************************************************************

//void int_gmp_real2int(ext_gmp_type r, int_gmp_type i)
//{
 //mpz_set_f(i,r);
 //return;
//}

//*****************************************************************************
// Converting GMP_integer to long integer
// added: 01.05.2001
// Last change: 01.05.2001
// Arguments: 1. GMP_integer
// 		  2. long integer
//*****************************************************************************

long int_gmp_int2long(int_gmp_type i)
{
 return mpz_get_si(i);
}


//*****************************************************************************
// SIMPLE ARITHMETIC OPERATIONS
//*****************************************************************************
// adding GMP_integers
// calculates z = z1 + z2
// added: 31.01.2001
// Last change: 31.01.2001
// Arguments: 1. GMP_integer
// 		  2. GMP_integer
// 		  3. GMP_integer
// ****************************************************************************

void int_gmp_add(int_gmp_type z1, int_gmp_type z2, int_gmp_type z)
{
 mpz_add(z,z1,z2);
 return;
}

//*****************************************************************************
// adding GMP_integer and unsigned long
// calculates z = z1 + z2
// added: 18.06.2001
// Last change: 18.06.2001
// Arguments: 1. GMP_integer
// 		  2. unsigned long int
// 		  3. GMP_integer
// ****************************************************************************

void int_gmp_add_ui(int_gmp_type z1, unsigned long z2, int_gmp_type z)
{
 mpz_add_ui(z,z1,z2);
 return;
}


//*****************************************************************************
// subtract GMP_integers
// calculates z = z1 - z2
// added: 01.02.2001
// Last change: 01.02.2001
// Arguments: 1. GMP_integer
// 		  2. GMP_integer
// 		  3. GMP_integer
// ****************************************************************************

void int_gmp_sub(int_gmp_type z1,int_gmp_type z2,int_gmp_type z)
{
  mpz_sub(z,z1,z2);
  return;
}

//*****************************************************************************
// subtract GMP_integer and unsigned long
// calculates z = z1 - z2
// added: 01.02.2001
// Last change: 01.02.2001
// Arguments: 1. GMP_integer
// 		  2. unsigned long
// 		  3. GMP_integer
// ****************************************************************************

void int_gmp_sub_ui(int_gmp_type z1,unsigned long z2,int_gmp_type z)
{
  mpz_sub_ui(z,z1,z2);
  return;
}


//*****************************************************************************
// multiplicate GMP-integers
// calculates z = z1 * z2
// added: 01.02.2001
// Last change: 01.02.2001
// Arguments: 1. GMP_integer
// 		  2. GMP_integer
// 		  3. GMP_integer
// ****************************************************************************

void int_gmp_mul(int_gmp_type z1, int_gmp_type z2, int_gmp_type z)
{
 mpz_mul(z,z1,z2);
 return;
}

//*****************************************************************************
// multiplicate GMP-integer by signed long
// calculates z = z1 * z2
// added: 01.02.2001
// Last change: 01.02.2001
// Arguments: 1. GMP_integer
// 		  2. signed long
// 		  3. GMP_integer
// ****************************************************************************

void int_gmp_mul_si(int_gmp_type z1, signed long z2, int_gmp_type z)
{
 mpz_mul_si(z,z1,z2);
 return;
}


//*****************************************************************************
// dividing GMP_integers
// calculates z = z1 / z2 == z1 DIV z2
// added: 01.02.2001
// Last change: 01.02.2001
// Arguments: 1. GMP_integer
// 		  2. GMP_integer
// 		  3. GMP_integer
// ****************************************************************************

void int_gmp_div(int_gmp_type z1, int_gmp_type z2, int_gmp_type z)
{
 mpz_tdiv_q(z,z1,z2); //**** truncated div ****
 return;
}

//*****************************************************************************
// dividing GMP_integer by unsigned long
// calculates z = z1 / z2 == z1 DIV z2
// added: 18.06.2001
// Last change: 18.06.2001
// Arguments: 1. GMP_integer
// 		  2. unsigned long
// 		  3. GMP_integer
// ****************************************************************************

void int_gmp_div_ui(int_gmp_type z1, unsigned long z2, int_gmp_type z)
{
 unsigned long dummy;
 dummy = mpz_tdiv_q_ui(z,z1,z2); //**** truncated div ****
 return;
}


//*****************************************************************************
// absolute value of integer in GMP
// returns z = |z1|
// added: 31.01.2001
// Last change: 31.01.2001
// Arguments: 1. GMP_integer
// 		  2. GMP_integer
//****************************************************************************

void int_gmp_abs(int_gmp_type z1, int_gmp_type z)
{
 mpz_abs(z, z1);
 return;
}

//****************************************************************************
// changes sign of GMP integer
// returns z = -z
// added: 18.06.2001
// Last change: 18.06.2001
// Arguments: 1. GMP_integer
// 		  2. GMP_integer
// Comment: used for operators with unsigned long integers
//****************************************************************************

void int_gmp_neg(int_gmp_type z1, int_gmp_type z)
{
  mpz_neg(z, z1);
  return;
}


//******************************************************************************
// nth root of integer in GMP
// added: 07.02.2001
// Last change: 07.02.2001
// Comment: this root-function truncates the result to the integer part
// Arguments: 1. GMP_integer
// 		  2. unsigned long int
// 		  3. GMP_integer
//******************************************************************************

void int_gmp_root(int_gmp_type z1, unsigned long int z2, int_gmp_type z)
{
 int temp;	
 temp=mpz_root(z, z1, z2);
}

//*******************************************************************************
// power function in GMP
// calculates z = z1 ^ z2
// added: 07.02.2001
// Last change: 07.02.2001
// Comment: base=MP exponent=long
// Arguments: 1. GMP_integer
// 		  2. unsigned long integer
// 		  3. GMP_integer
//*******************************************************************************

void int_gmp_power_i(int_gmp_type z1, unsigned long int z2, int_gmp_type z)
{
 mpz_pow_ui(z, z1, z2);
 return;
}

//*******************************************************************************
// power function in GMP
// calculates z = z1 ^ z2
// added: 07.02.2001
// Last change: 07.02.2001
// Comment: base=long exponent=long
// Arguments: 1. unsigned long int
// 		  2. unsigned long int
// 		  3. GMP_integer
//*******************************************************************************

void int_gmp_power_ii(unsigned long int z1, unsigned long int z2, int_gmp_type z)
{
 mpz_ui_pow_ui(z, z1, z2);
 return;
}

//*******************************************************************************
// faculty function in GMP
// calculates z = z1!
// added: 07.02.2001
// Last change: 07.02.2001
// Comment: base is long
// Arguments: 1. unsigned long
// 		  2. GMP_integer
//*******************************************************************************

void int_gmp_fac(unsigned long int z1, int_gmp_type z)
{
 mpz_fac_ui(z, z1);
 return;
}

//*******************************************************************************
// modulo function in GMP
// calculates z = z1 % z2
// added: 07.02.2001
// Last change: 07.02.2001
// Arguments: 1. GMP_integer
// 		  2. GMP_integer
// 		  3. GMP_integer
//*******************************************************************************

void int_gmp_modulo(int_gmp_type z1, int_gmp_type z2, int_gmp_type z)
{
 mpz_mod(z, z1, z2);
 return;
}

//*****************************************************************************
// logarithm with base 2 
// calculates log(z)
// added: 31.03.2001
// Last change: 31.03.2001
// Arguments: 1. GMP_integer
// 		  return: long integer 
//*****************************************************************************

long int_gmp_log(int_gmp_type z)
{
unsigned long r=0,w,n,m;
if (z->_mp_size==0) return GMP_min;
w=(z->_mp_d)[abs(z->_mp_size)-1];   //****** get the highest-valued limb *****
n=w;
for (m=mp_bits_per_limb>>1; m>0; m=m>>1 ) //****** divide and conquer ******
	{
      if (n>= (1<<m) )
		{
        	r+=m; n=n>>m;
      	}
    	}
if (w > (1<<r) || abs(z->_mp_size) >0) r++;    //****** 
return r;
}

//****************************************************************************************

long int_gmp_log2(int_gmp_type z)
{
	unsigned long r=0,n,m;
	mp_limb_t temp;// could be long
	temp=mpz_getlimbn(z,1);
	n=temp;
	for (m=mp_bits_per_limb>>1; m>0; m=m>>1 ) //****** kinda divide and conquer ******
		{
      	if (n>= (1<<m) )
		{
        		r+=m; n=n>>m;
      	}
    	}
	if (temp > (1<<r) || abs(z->_mp_size) >0) r++;    //****** ungerade ******
	return r+(abs(z->_mp_size)-1)*32-1;
}

//*****************************************************************************
// output of integer
// writes z to STDOUT with length w
// added: 31.01.2001
// Last change: 31.01.2001
// Arguments: 1. GMP_integer
// 		  2. long integer
//****************************************************************************


void int_gmp_writee(int_gmp_type z, long w)
{
  char *s;
  long l;
  // char s[mpz_sizeinbase(z,10)+2];
  s=mpz_get_str(NULL, 10, z);
  l=w-strlen(s);
  for(;l>0;l--)printf(" ");
  printf("%s", &s[0]);
  free(s);
}

void int_gmp_outstr(int_gmp_type z, long w) 
{
  mpz_out_str(stdout,10,z);
}


//*****************************************************************************
// writes integer to a string
// writes z to string with length w
// added: 18.07.2001
// Last change: 18.07.2001
// Arguments: 1. GMP_integer
// 		  2. long integer
//	return: char*
//****************************************************************************


char* int_gmp_swritee(int_gmp_type z, long w)
{
  	char *s=malloc(w),*r;
  	long l;
  	r=mpz_get_str(NULL, 10, z);
  	l=w-strlen(r);
  	if(l>0)
		{
		char *temp=malloc(l+1);
		temp[l]='\0';
		for(;l>0;l--)temp[l-1]=' ';
		sprintf(s,"%s%s",temp,r);
		free(temp);
		}
	free(r);
  	return s;
}


//*****************************************************************************
// output of integer
// writes z to STDOUT with length calculated
// added: 28.02.2001
// Last change: 28.02.2001
// Arguments: 1. GMP_integer
//****************************************************************************

void int_gmp_printf(int_gmp_type z)
{
  char s[mpz_sizeinbase(z,10)+2];
  mpz_get_str(s,10,z);
  if (s[0]=='-') {printf("-%s", &s[1]);} else printf("%s", &s[0]);
}

//*****************************************************************************
// writes z to string with length calculated
// added: 18.07.2001
// Last change: 18.07.2001
// Arguments: 1. GMP_integer
// 	return: char*
//****************************************************************************

char* int_gmp_sprintf(int_gmp_type z)
{
  char* result=mpz_get_str(NULL,10,z);
  return result;
}


//*****************************************************************************
// output integer in GMP
// added: 01.02.2001
// Last change: 01.02.2001
// writes limb per limb
//****************************************************************************

void int_gmp_write(int_gmp_type z, long w)
{
  char s[w];
  long i;
  mpz_get_str(s, 10, z);
  if (z->_mp_size != 0) {
     printf("\n");
     for (i=abs(z->_mp_size-1);i>=0; i-=1) 
       printf("%lu ",z->_mp_d[i]);
  }
//printf("\nLimbs: %d, Bits: %ld\n",,z->_mp_size,mpz_get_prec(z));
}



//*****************************************************************************
// duplicate initialized integer in GMP
// copies z1 into already initialized GMP_integer z2
// added: 01.02.2001
// Last change: 01.02.2001
// Arguments: 1. GMP_integer
// 		  2. GMP_integer
//****************************************************************************

void int_gmp_duplicate_wo_init(int_gmp_type z1, int_gmp_type z2)
{
 mpz_set(z2,z1);
 return;
}



//*****************************************************************************
// duplicate non-initialized integer in GMP
// copies z1 into non-initialized GMP_integer z2
// added: 01.02.2001
// Last change: 01.02.2001
// Arguments: 1. GMP_integer
// 		  2. GMP_integer
//****************************************************************************

void int_gmp_duplicate_w_init(int_gmp_type z1, int_gmp_type *z2)
{
  *z2=int_gmp_init();
  mpz_set(*z2,z1);
}


//*****************************************************************************
// square root of GMP_integer
// calculates z = squareroot(z1)
// added: 01.02.2001
// Last change: 01.02.2001
// Arguments: 1. GMP_integer
// 		  2. GMP_integer
//****************************************************************************

void int_gmp_sqrt(int_gmp_type z1, int_gmp_type z)
{
 mpz_sqrt(z,z1);
 return;
}


//*****************************************************************************
// shifting GMP_integer by n bits
// calculates z = z1 <<(>>) z2
// added: 01.02.2001
// Last change: 01.02.2001
// Arguments: 1. GMP_integer
// 		  2. GMP_integer
// 		  3. long integer n
//****************************************************************************

void int_gmp_shift(int_gmp_type z1, int_gmp_type z, long n)
{
  if (n>=0) mpz_mul_2exp(z, z1, n);
  else mpz_tdiv_q_2exp(z, z1, -n);  //** truncate-div **
}


//*****************************************************************************
// COMPARING FUNCTIONS
//*****************************************************************************
// Comparing 2 MP_integers and returning integer as result:
// values: -1 (less than), 0 (equal), 1 (greater than)
// added: 01.05.2001
// Last change: 01.05.2001
// Arguments: 1. MP_integer
// 		  2. MP_integer
// 		  return: (standard) integer
//*****************************************************************************

int int_gmp_cmp(int_gmp_type z1, int_gmp_type z2)
{
return mpz_cmp(z1,z2);
}

//*****************************************************************************
// Return sign of MP_integer as (standard) integer:
// values: -1 (negative), 0 (zero), 1 (positive)
// added: 01.05.2001
// Last change: 01.05.2001
// Arguments: 1. MP_integer
// 		  2. MP_integer
// 		  return: (standard) integer
//*****************************************************************************

int int_gmp_sgn(int_gmp_type z1)
{
return mpz_sgn(z1);
}
