/*

LRGMP_rat_ext.c -- extension to the GMP rational functions for the 
                   iRRAM library with backend LRGMP
 
Copyright (C) 2001 Norbert Mueller
 
This file is part of the iRRAM Library.
 
The iRRAM Library is free software; you can redistribute it and/or modify
it under the terms of the GNU Library General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at your
option) any later version.
 
The iRRAM Library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
License for more details.
 
You should have received a copy of the GNU Library General Public License
along with the iRRAM Library; see the file COPYING.LIB.  If not, write to
the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
MA 02111-1307, USA. 
*/

#include "LRGMP_rat_ext.h"
#include "LRGMP_int_ext.h"

#include <strings.h>

#ifndef DEBUG
#define DEBUG 0
#endif
#if DEBUG > 0
  long DEBUG_OP = 1;
  unsigned long op_count=0;
  long minshowop = DEBUG ;
#endif


#define MaxFreeVars 1000
static rat_gmp_type gmp_FreeVarsi[MaxFreeVars];
static long gmp_FreeVarCount=0L;
long rat_gmp_var_count=0;

void rat_gmp_initialize()
{
/*not necessary*/
}


/*****************************************************************************/
// initialize rational in GMP, allocating memory
// added: 19.02.2001
// Last change: 19.02.2001
/*****************************************************************************/

rat_gmp_type rat_gmp_init()
{
	rat_gmp_type z;
	if (gmp_FreeVarCount>0) 
		{
		gmp_FreeVarCount--;
		z = gmp_FreeVarsi[gmp_FreeVarCount];
		}
	else
		{
		z = (rat_gmp_type) malloc(sizeof(__mpq_struct));
		mpq_init(z);
		}
	rat_gmp_var_count++;
	return z;
}





/*****************************************************************************/
// delete no longer used rationals's, including deallocation
// added: 19.02.2001
// Last change: 19.02.2001
/*****************************************************************************/

void rat_gmp_free(rat_gmp_type z)
{
	if(gmp_FreeVarCount < MaxFreeVars)
		{
		gmp_FreeVarsi[gmp_FreeVarCount]=z;
		gmp_FreeVarCount++;
		}
	else
		{
		mpq_clear(z);
		free(z);
		}
	rat_gmp_var_count--;
}

/*****************************************************************************/
// canonicalize the given rational, used for used defined rationals
// added: 19.02.2001
// Last change: 19.02.2001
/*****************************************************************************/
void rat_gmp_canon(rat_gmp_type z)
{
	mpq_canonicalize(z);
}

/*****************************************************************************/
// Getting size of MP rationals
// til now for debugging reasons only 
// returns size(num)+size(denum)
/*****************************************************************************/

int rat_gmp_getsize(rat_gmp_type z)
{
	int_gmp_type num;
	int_gmp_type den;
	num=int_gmp_init();
	den=int_gmp_init();
	mpq_get_num(num,z);
	mpq_get_den(den,z);

  	//return num->_mp_size+den->_mp_size;
	return int_gmp_getsize(num)+int_gmp_getsize(den);

	int_gmp_free(num);
	int_gmp_free(den);
}

/*****************************************************************************/
// Getting binary size of MP rationals
// used for setting precision while converting to REAL
// returns binsize(num)+binsize(denum)
/*****************************************************************************/

long rat_gmp_binsize(rat_gmp_type z)
{
	int_gmp_type num;
	int_gmp_type den;
	num=int_gmp_init();
	den=int_gmp_init();
	mpq_get_num(num,z);
	mpq_get_den(den,z);
	
	return mpz_sizeinbase(num,2)+mpz_sizeinbase(den,2);

	int_gmp_free(num);
	int_gmp_free(den);
}

/*****************************************************************************/
// adding rational in GMP
// added: 19.02.2001
// Last change: 19.02.2001
/****************************************************************************/

void rat_gmp_add(rat_gmp_type z1, rat_gmp_type z2, rat_gmp_type z)
{
	mpq_add(z,z1,z2);
return;
}

/*****************************************************************************/
// adding rational and long in GMP
// added: 18.06.2001
// Last change: 18.06.2001
/*****************************************************************************/

void rat_gmp_add_ui(rat_gmp_type z1, unsigned long z2, rat_gmp_type z)
{
#if DEBUG>0
#endif
	int_gmp_type num;
	int_gmp_type den;
	int_gmp_type temp;
	num=int_gmp_init();
	den=int_gmp_init();
	temp=int_gmp_init();
	mpq_get_num(num,z1);
	mpq_get_den(den,z1);

	mpz_mul_ui(temp,den,z2);
	mpz_add(num,num,temp);

	mpq_set_num(z,num);
	mpq_set_den(z,den); 
	mpq_canonicalize(z);
	int_gmp_free(num);
	int_gmp_free(den);
	int_gmp_free(temp);

return;
}


/*****************************************************************************/
// subtract rational in GMP
// added: 19.02.2001
// Last change: 19.02.2001
/****************************************************************************/

void rat_gmp_sub(rat_gmp_type z1,rat_gmp_type z2,rat_gmp_type z)
{
  	mpq_sub(z,z1,z2);
  return;
}

/*****************************************************************************/
// subtract rational and long in GMP
// added: 18.06.2001
// Last change: 18.06.2001
/*****************************************************************************/

void rat_gmp_sub_ui(rat_gmp_type z1, unsigned long z2, rat_gmp_type z)
{
#if DEBUG>0
#endif
	int_gmp_type num;
	int_gmp_type den;
	int_gmp_type temp;
	num=int_gmp_init();
	den=int_gmp_init();
	temp=int_gmp_init();
	mpq_get_num(num,z1);
	mpq_get_den(den,z1);

	mpz_mul_ui(temp,den,z2);
	mpz_sub(num,num,temp);

	mpq_set_num(z,num);
	mpq_set_den(z,den); 
	mpq_canonicalize(z);
	int_gmp_free(num);
	int_gmp_free(den);
	int_gmp_free(temp);

#if (DEBUG>0)
#endif

return;
}


/*****************************************************************************/
// multiplicate rational in GMP
// added: 19.02.2001
// Last change: 19.02.2001
/*****************************************************************************/

void rat_gmp_mul(rat_gmp_type z1, rat_gmp_type z2, rat_gmp_type z)
{
  	mpq_mul(z,z1,z2);
  return;
}

/*****************************************************************************/
// multiplicate rational and long in GMP
// added: 18.06.2001
// Last change: 18.06.2001
/*****************************************************************************/

void rat_gmp_mul_si(rat_gmp_type z1, unsigned long z2, rat_gmp_type z)
{
	int_gmp_type num;
	int_gmp_type den;
	num=int_gmp_init();
	den=int_gmp_init();
	mpq_get_num(num,z1);
	mpq_get_den(den,z1);

	mpz_mul_si(num,num,z2);

	mpq_set_num(z,num);
	mpq_set_den(z,den); 
	mpq_canonicalize(z);
	int_gmp_free(num);
	int_gmp_free(den);

return;
}


/*****************************************************************************/
// dividing rational in GMP
// added: 19.02.2001
// Last change: 19.02.2001
/****************************************************************************/

void rat_gmp_div(rat_gmp_type z1, rat_gmp_type z2, rat_gmp_type z)
{
  	mpq_div(z,z1,z2); //**** truncated div ****
}

/*****************************************************************************/
// dividing rational and long in GMP
// added: 18.06.2001
// Last change: 18.06.2001
/*****************************************************************************/

void rat_gmp_div_si(rat_gmp_type z1, unsigned long z2, rat_gmp_type z)
{
	int_gmp_type num;
	int_gmp_type den;
	num=int_gmp_init();
	den=int_gmp_init();
	mpq_get_num(num,z1);
	mpq_get_den(den,z1);

	mpz_mul_si(den,den,z2);

	mpq_set_num(z,num);
	mpq_set_den(z,den); 
	mpq_canonicalize(z);
	int_gmp_free(num);
	int_gmp_free(den);

#if (DEBUG>0)
#endif

return;
}

/*****************************************************************************/
// dividing rational and long in GMP
// added: 18.06.2001
// Last change: 18.06.2001
// ****************************************************************************

void rat_gmp_si_div(unsigned long z1, rat_gmp_type z2, rat_gmp_type z)
{
#if DEBUG>0
#endif
	int_gmp_type num;
	int_gmp_type den;
	num=int_gmp_init();
	den=int_gmp_init();
	mpq_get_num(num,z2);
	mpq_get_den(den,z2);

	mpz_mul_si(den,den,z1);

	mpq_set_num(z,den);	// invert
	mpq_set_den(z,num); 
	mpq_canonicalize(z);
	int_gmp_free(num);
	int_gmp_free(den);

#if (DEBUG>0)
#endif

return;
}



/*****************************************************************************/
// absolute value of rational in GMP
// added: 19.02.2001
// Last change: 19.02.2001
//****************************************************************************

void rat_gmp_abs(rat_gmp_type z1, rat_gmp_type z)
{

	int_gmp_type num;
	int_gmp_type den;
	num=int_gmp_init();
	den=int_gmp_init();
	mpq_get_num(num,z1);
	mpq_get_den(den,z1);
	mpz_abs(num, num);
	mpq_set_num(z,num);
	mpq_set_den(z,den); 
	mpq_canonicalize(z);
	int_gmp_free(num);
	int_gmp_free(den);
}


/*****************************************************************************/
// power function in GMP
// added: 19.02.2001
// Last change: 19.02.2001
// Comment: base=MP exponent=long
/*****************************************************************************/

void rat_gmp_power(rat_gmp_type z1, unsigned long int z2, rat_gmp_type z)
{
	int sign;
	int_gmp_type num;
	int_gmp_type den;
	int_gmp_type temp;
	num=int_gmp_init();
	den=int_gmp_init();
	temp=int_gmp_init();
	mpq_get_num(num,z1);
	mpq_get_den(den,z1);
	if (mpz_sgn(num)<0) sign=-1; else sign=1;
	mpz_set_si(temp,sign);
	if (sign==-1)mpz_mul(num,num,temp); //***** Nachprfen
	mpz_pow_ui(num,num,z2);
	mpz_pow_ui(den,den,z2);
	if (sign==-1)mpz_mul(num,num,temp);
	mpq_set_num(z,num);
	mpq_set_den(z,den);
	mpq_canonicalize(z);
	int_gmp_free(num);
	int_gmp_free(den);
}


/*****************************************************************************/
// Function to compute rational ^ rational
// added: 26.02.2001
// last change: 26.02.2001
// problems: int_gmp_power and int_gmp_root
/*****************************************************************************/

void rat_gmp_powerr(rat_gmp_type z1, rat_gmp_type z2, rat_gmp_type z)
{
	int_gmp_type num,num2;
	int_gmp_type den,den2;
	num=int_gmp_init();
	den=int_gmp_init();
	num2=int_gmp_init();
	den2=int_gmp_init();
	mpq_get_num(num,z1);
	mpq_get_den(den,z1);
	mpq_get_num(num2,z2);
	mpq_get_den(den2,z2);
	//int_gmp_power(num,num2,num);	//****** ???
	//int_gmp_root(num,den2,num);		//****** ???
	//int_gmp_power(den,num2,den);	//****** ???
	//int_gmp_root(den,den2,den);		//****** ???
	mpq_set_num(z,num);
	mpq_set_den(z,den);
	mpq_canonicalize(z);
	int_gmp_free(num);
	int_gmp_free(den);
	int_gmp_free(num2);
	int_gmp_free(den2);
}




/*****************************************************************************/
// needed for log2
/*****************************************************************************/

long rat_gmp_log(rat_gmp_type z)
{
	int_gmp_type num;
	int_gmp_type den;
	unsigned long rn=0,rd=0,w,n,m;
	num=int_gmp_init(); 
	den=int_gmp_init();
	mpq_get_num(num,z);
	mpq_get_den(den,z);

	if (num->_mp_size==0) rn=GMP_min;
	w=(num->_mp_d)[abs(num->_mp_size)-1];   //****** get the highest-valued limb *****
	n=w;
	for (m=mp_bits_per_limb>>1; m>0; m=m>>1 ) //****** divide and conquer ******
		{
	      if (n>= (1<<m) )
			{
        		rn+=m; n=n>>m;
      		}
    		}
	if (w > (1<<rn) || abs(num->_mp_size) >0) rn++;    //****** 

	if (den->_mp_size==0) rd=1;
	w=(den->_mp_d)[abs(den->_mp_size)-1];   //****** get the highest-valued limb *****
	n=w;
	for (m=mp_bits_per_limb>>1; m>0; m=m>>1 ) //****** divide and conquer ******
		{
      	if (n>= (1<<m) )
			{
        		rd+=m; n=n>>m;
      		}
    		}
	if (w > (1<<rd) || abs(den->_mp_size) >0) rd++;    //****** 

	int_gmp_free(num);
	int_gmp_free(den);

	return rn-rd;
}



/*****************************************************************************/
// output of rational
// added: 19.02.2001
// Last change: 19.02.2001
//****************************************************************************


void rat_gmp_writee(rat_gmp_type z, long w)
{
	int_gmp_type num;
	int_gmp_type den;
  	char *n, *d;
	long l;
	num=int_gmp_init();
	den=int_gmp_init();
	mpq_get_num(num,z);
	mpq_get_den(den,z);

  	n=mpz_get_str(NULL, 10, num);
	d=mpz_get_str(NULL, 10, den);
	l=w-strlen(d)-strlen(n)-1;
	for(;l>0;l--)printf(" ");
	printf("%s/%s", &n[0], &d[0]);
	int_gmp_free(num);
	int_gmp_free(den);
	free(n);free(d);
}

void rat_gmp_outstr(rat_gmp_type z, long w) 
{
  mpq_out_str(stdout,10,z);
}

/*****************************************************************************/
// writes rational to a string
// added: 18.07.2001
// Last change: 18.07.2001
/****************************************************************************/


char* rat_gmp_swritee(rat_gmp_type z, long w)
{
	int_gmp_type num;
	int_gmp_type den;
  	char *n, *d, *s=malloc(w), *t;
	long l;

	num=int_gmp_init();
	den=int_gmp_init();
	mpq_get_num(num,z);
	mpq_get_den(den,z);

	n=mpz_get_str(NULL, 10, num);
	d=mpz_get_str(NULL, 10, den);
	l=w-strlen(n)-strlen(d)-1;
	t=malloc(l);
  	if(l>0)
		{	
		char temp[l];
		temp[l]='\0';
		for(;l>0;l--)temp[l-1]=' ';
		sprintf(t,"%s",temp);
		}

	int_gmp_free(num);
	int_gmp_free(den);
	sprintf(s,"%s%s/%s",t,n,d);
	free(t);free(n);free(d);
	return s;
}



void rat_gmp_printf(rat_gmp_type z)
{
	int_gmp_type num=int_gmp_init();
	int_gmp_type den=int_gmp_init();
  	char n[mpz_sizeinbase(num,10)+2], d[mpz_sizeinbase(den,10)+2];
	
	
	mpq_get_num(num,z);
	mpq_get_den(den,z);

  	mpz_get_str(n, 10, num);
	mpz_get_str(d, 10, den);
  	if (n[0]=='-') {printf("-%s/%s", &n[1],&d[0]);} else printf("%s/%s", &n[0], &d[0]);
	int_gmp_free(num);
	int_gmp_free(den);

}

char* rat_gmp_sprintf(rat_gmp_type z)
{
	int_gmp_type num=int_gmp_init();
	int_gmp_type den=int_gmp_init();
  	char *n, *d;
	char *s;	
	
	mpq_get_num(num,z);
	mpq_get_den(den,z);

  	n=mpz_get_str(NULL, 10, num);
	d=mpz_get_str(NULL, 10, den);
	s=malloc(strlen(n)+strlen(d)+2);
	//s=strcat(&n,t);
	//s=strcat(s,&d);
	sprintf(s,"%s/%s",n,d);
	int_gmp_free(num);
	int_gmp_free(den);
	return s;
}


/*****************************************************************************/
// duplicate initialized rational in GMP
// added: 19.02.2001
// Last change: 19.02.2001
// Sinn?
//****************************************************************************

void rat_gmp_duplicate_wo_init(rat_gmp_type z1, rat_gmp_type z2)
{
  mpq_set(z2,z1);
}



/*****************************************************************************/
// duplicate non-initialized rational in GMP
// added: 19.02.2001
// Last change: 19.02.2001
// Sinn?
//****************************************************************************

void rat_gmp_duplicate_w_init(rat_gmp_type z1, rat_gmp_type *z2)
{
  *z2=rat_gmp_init();
  mpq_set(*z2,z1);
}



/*****************************************************************************/
// duplicate non-initialized rational in GMP
// added: 19.02.2001
// Last change: 19.02.2001
// ?correct?
//****************************************************************************

void rat_gmp_shift(rat_gmp_type z1, rat_gmp_type z, long n)
{
	int_gmp_type num,den;
	num=int_gmp_init();
	den=int_gmp_init();
	mpq_get_num(num,z1);
	mpq_get_den(den,z1);

  	if (n>=0) mpz_mul_2exp(num, num, n);
  	//else mpz_tdiv_q_2exp(num, num, -n);  //****** auf den anwenden
	else mpz_mul_2exp(den, den, -n);	

	mpq_set_num(z,num);
	mpq_set_den(z,den);
	mpq_canonicalize(z);

	int_gmp_free(num);
	int_gmp_free(den);
}

/*****************************************************************************/
// convert string to rational
// Added: 26.02.2001
// Last change: 28.02.2001
/*****************************************************************************/

void rat_gmp_string_2_rat(rat_gmp_type z, char* s)
{
	int_gmp_type num,den;
	int num1, den1;
	num=int_gmp_init();
	den=int_gmp_init();
	sscanf(s,"%d/%d",&num1,&den1);
	mpz_set_si(num,num1);
	mpz_set_si(den,den1);
	mpq_set_num(z,num);
	mpq_set_den(z,den);
	mpq_canonicalize(z);
	int_gmp_free(num);
	int_gmp_free(den);
}

/*****************************************************************************/
// Convert double to rational
// Added: 26.02.2001
// Last change: 28.02.2001
// not so good
/*****************************************************************************/

void rat_gmp_double_2_rat(rat_gmp_type z, double d)
{
	int_gmp_type num,den;
	char* s='\0';
	int num1=0, den1=1;
	num=int_gmp_init();
	den=int_gmp_init();
	sprintf(s,"%10.10f",d); 
	sscanf(s,"%d.%d",&num1,&den1);
	mpz_set_si(num,(10^10)*num1+den1);
	mpz_set_si(den,10^10);
	mpq_set_num(z,num);
	mpq_set_den(z,den);
	mpq_canonicalize(z);
	int_gmp_free(num);
	int_gmp_free(den);

}

/*****************************************************************************/
// Return canonical numerator of rational number
/*****************************************************************************/

void rat_gmp_get_numerator(int_gmp_type z, rat_gmp_type z1)
{
	mpq_get_num(z,z1);
}


/*****************************************************************************/
// Return canonical denominator of rational number
/*****************************************************************************/

void rat_gmt_get_denominator(int_gmp_type z, rat_gmp_type z1)
{
	mpq_get_den(z,z1);
}
