/*

Integer.cc -- routines for INTEGER data type
 
Copyright (C) 2001-2003 Norbert Mueller, Tom van Diessen
 
This file is part of the iRRAM Library.
 
The iRRAM Library is free software; you can redistribute it and/or modify
it under the terms of the GNU Library General Public License as published by
the Free Software Foundation; either version 2 of the License, or (at your
option) any later version.
 
The iRRAM Library is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Library General Public
License for more details.
 
You should have received a copy of the GNU Library General Public License
along with the iRRAM Library; see the file COPYING.LIB.  If not, write to
the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston,
MA 02111-1307, USA. 
*/


/*
Changelog: (initial version by Tom, all modifications by Norbert)

  2003-10-10  Modifications concerning automatic type conversions 

  2001-08-20  Modifications in the inheritance: 
              The numbers are no longer METRIC_OBJECTs

  2001-07-19  Initial version from Tom's diploma thesis

*/

#include <stdlib.h>
#include <stdarg.h>
#include <string.h>

#include "iRRAM_core.h"

#define maxslong 2147483647
#define maxlong 2*(maxslong-1) //****** must be unsigned ******

//***************************************************************************************
// Variables for storing the values during limit-calculations
//***************************************************************************************

unsigned long cachesize_ikonvert=10;
unsigned long ikonvertanzahl=0;
unsigned long ikonvertaufrufe=0;
MP_int_type * ikonvert;

//****************************************************************************************
// CONSTRUCTORS
// IMPORTANT: Objects MUST be initialized before being used.
//****************************************************************************************
//****************************************************************************************

INTEGER::INTEGER(MP_int_type y){
  value = y;
}

INTEGER::INTEGER(){
  MP_int_init(value);
  MP_int_to_INTEGER(0,value);
}

INTEGER::INTEGER(const int i){
  MP_int_init(value);
  MP_int_to_INTEGER(i,value);
}

INTEGER::INTEGER(const long i){
  MP_int_init(value);
  MP_int_to_INTEGER(i,value);
}

INTEGER::INTEGER(const double d){
  MP_int_init(value);
  MP_double_to_INTEGER(d,value);
}

INTEGER::INTEGER(const INTEGER& i){
  MP_int_init(value);
  MP_int_duplicate_wo_init(i.value, value);
}


//****************************************************************************************
// Constructing INTEGER from character string
// The base is set to 10 (decimal)
//****************************************************************************************

INTEGER::INTEGER(const char* s){
  MP_int_init(value);
  MP_string_to_INTEGER(s,value,10);
}


//****************************************************************************************
// Copy constructor = from existing INTEGER-object
//****************************************************************************************

INTEGER& INTEGER::operator = (const INTEGER& y)
{
  MP_int_duplicate_wo_init(y.value, value);
  return (*this);
}

INTEGER& INTEGER::operator = (const int y)
{
  MP_int_to_INTEGER(y,value);
  return (*this);
}

INTEGER& INTEGER::operator = (const long y)
{
  MP_int_to_INTEGER(y,value);
  return (*this);
}

//****************************************************************************************
// Standard Destructor
//****************************************************************************************

INTEGER::~INTEGER()
{
  MP_int_clear(value);
  value=NULL;
}


//****************************************************************************************
// Returns the size of INTEGER-objects in number of limbs
//****************************************************************************************

int INTEGER::size()
{
  return MP_int_getsize(value);
}

//****************************************************************************************
// Shifting an INTEGER-value
//****************************************************************************************

#ifdef MP_shift

INTEGER scale(const INTEGER& x, const long n)
{
	MP_int_type zvalue;
	MP_int_init(zvalue);
	MP_int_shift(x.value,zvalue,n);
	return zvalue;
}

#else

INTEGER scale(const INTEGER& x, const long n)
{
	if (n==0) return x;
	if (n==1) return 2*x;
	INTEGER y=1;
	INTEGER xc=2;
	if (n<0) {xc=y/2;n=-n;}
	if (n==1) return xc*x;
	for (long k=n;k>0;k=k/2)
	{
		if (k%2==1) y*=xc;
		if (k==1) break;
		xc*=xc;
	}
	return y*x;
}

#endif


//****************************************************************************************
// Power: return  = x ^ n
//****************************************************************************************

INTEGER operator ^ (const INTEGER& x, const unsigned long n)
{
  MP_int_type zvalue;
  MP_int_init(zvalue);
  MP_int_power_i(x.value,n,zvalue);
  return zvalue;
}

//****************************************************************************************
// Power: return = x ^ y
//****************************************************************************************

INTEGER operator ^ (const INTEGER& x, const INTEGER& y)
{
   if (y==0 || x==1) return 1;
   if (y==1) return x;
   if(y>maxlong) throw iRRAM_Numerical_Exception(iRRAM_integer_error);
   return x^(unsigned long)(MP_INTEGER_to_long(y.value));
}


//****************************************************************************************
// Shifter: returns x << n
// shifts n bits to the left
//****************************************************************************************

INTEGER operator << (const INTEGER& x, const long n){
  return scale(x,n);
}

//****************************************************************************************
// Shifter: returns x >> n
// shifts n bits to the right
//****************************************************************************************

INTEGER operator >> (const INTEGER& x, const long n){
  return scale(x,-n);
}

//****************************************************************************************
// Addition: returns x + y
//****************************************************************************************

INTEGER operator + (const INTEGER& x, const INTEGER& y){
  MP_int_type zvalue;
  MP_int_init(zvalue);
  MP_int_add(x.value,y.value,zvalue);
  return zvalue;
}

INTEGER operator + (const INTEGER& x, const long y){
  MP_int_type zvalue;
  MP_int_init(zvalue);
  if (y<0) MP_int_sub_ui(x.value,-y,zvalue);
  else     MP_int_add_ui(x.value,y,zvalue);
  return zvalue;
}


INTEGER operator + (const long     x, const INTEGER& y) {return y+x;}
INTEGER operator + (const INTEGER& x, const int      y) {return x+long(y);};
INTEGER operator + (const int      x, const INTEGER& y) {return long(x)+y;};


//****************************************************************************************
// Addition: returns x += y
//****************************************************************************************

INTEGER& operator += (INTEGER& x, const INTEGER& y){
  MP_int_type zvalue;
  MP_int_init(zvalue);
  MP_int_add(x.value,y.value,zvalue);
  MP_int_clear(x.value);
  x.value=zvalue;
  return x;
}

INTEGER& operator += (INTEGER& x, const long n){ return x=x+n; }
INTEGER& operator += (INTEGER& x, const int n ){ return x=x+n; }


//****************************************************************************************
// Subtraction: returns x - y
//****************************************************************************************

INTEGER operator - (const INTEGER& x, const INTEGER& y){
  MP_int_type zvalue;
  MP_int_init(zvalue);
  MP_int_sub(x.value,y.value,zvalue);
  return zvalue;
}

INTEGER operator - (const INTEGER& x, const long y){
  MP_int_type zvalue;
  MP_int_init(zvalue);
  if(y<0) MP_int_add_ui(x.value,-y,zvalue);
    else  MP_int_sub_ui(x.value,y,zvalue);
  return zvalue;
}

INTEGER operator - (const long x, const INTEGER& y){
  return -(y-x);
}

INTEGER operator - (const INTEGER& x, const int      y) {return x-long(y);};
INTEGER operator - (const int      x, const INTEGER& y) {return long(x)-y;};


//****************************************************************************************
// Subtraction: returns x -= y
//****************************************************************************************

INTEGER& operator -= (INTEGER& x, const INTEGER& y){
  MP_int_type zvalue;
  MP_int_init(zvalue);
  MP_int_sub(x.value,y.value,zvalue);
  MP_int_clear(x.value);
  x.value=zvalue;
  return x;
}

INTEGER& operator -= (INTEGER& x, const long n){ return x=x-n; }
INTEGER& operator -= (INTEGER& x, const int  n){ return x=x-n; }

//****************************************************************************************
// Negative: returns -x
//****************************************************************************************

INTEGER operator - (const INTEGER& x){
  MP_int_type zvalue;
  INTEGER y=INTEGER(long(0));
  MP_int_init(zvalue);
  MP_int_sub(y.value,x.value,zvalue);
  return zvalue;
}

//****************************************************************************************
// Multiplication: returns x * y
//****************************************************************************************

INTEGER operator * (const INTEGER& x, const INTEGER& y){
  MP_int_type zvalue;
  MP_int_init(zvalue);
  MP_int_mul(x.value,y.value,zvalue);
  return zvalue;
}

INTEGER operator * (const INTEGER& x, const long y){
  MP_int_type zvalue;
  MP_int_init(zvalue);
  MP_int_mul_si(x.value,y,zvalue);
  return(zvalue);
}

INTEGER operator * (const long x, const INTEGER& y){
  MP_int_type zvalue;
  MP_int_init(zvalue);
  MP_int_mul_si(y.value,x,zvalue);
  return(zvalue);
}

INTEGER operator * (const INTEGER& x, const int      y) {return x*INTEGER(y);};
INTEGER operator * (const int      x, const INTEGER& y) {return INTEGER(x)*y;};

//****************************************************************************************
// Multiplication: returns x *= y
//****************************************************************************************

INTEGER& operator *= (INTEGER& x, const INTEGER& y){
  MP_int_type zvalue;
  MP_int_init(zvalue);
  MP_int_mul(x.value,y.value,zvalue);
  MP_int_clear(x.value);
  x=zvalue;
  return x;
}

INTEGER& operator *= (INTEGER& x, const long n){ return x=x*n; }
INTEGER& operator *= (INTEGER& x, const int n ){ return x=x*n; }

//****************************************************************************************
// Division: returns x / y
//****************************************************************************************


INTEGER operator / (const INTEGER& x, const INTEGER& y){
  MP_int_type zvalue;
  MP_int_init(zvalue);
  MP_int_div(x.value,y.value,zvalue);
  return zvalue;
}


INTEGER operator / (const INTEGER& x, const long n){
  MP_int_type zvalue;
  MP_int_init(zvalue);
  if(n<0){
    MP_int_div_ui(x.value,-n,zvalue);
    MP_int_mul_si(zvalue,-1,zvalue);
  }
  else MP_int_div_ui(x.value,n,zvalue);
  return zvalue;
}

INTEGER operator / (const INTEGER& x, const int      y) {return x/long(y);};
INTEGER operator / (const int      x, const INTEGER& y) {return INTEGER(x)/y;};
INTEGER operator / (const long     x, const INTEGER& y) {return INTEGER(x)/y;};

//****************************************************************************************
// Modulo: returns x % y
//****************************************************************************************

INTEGER operator % (const INTEGER& x, const INTEGER& y){
  MP_int_type zvalue;
  MP_int_init(zvalue);
  MP_int_modulo(x.value,y.value,zvalue);
  return zvalue;
}


//****************************************************************************************
// Square-root: returns SQRT(x)
// Returns the truncated square-root of the given argument
//****************************************************************************************

INTEGER sqrt (const INTEGER& x)
{
	MP_int_type zvalue;
	MP_int_init(zvalue);
	MP_int_sqrt(x.value, zvalue);
	return zvalue;
}


//****************************************************************************************
// Logarithm to base 2
// returns the truncated logarithm with base 2
// 1. argument: INTEGER
// return value: INTEGER
//****************************************************************************************

INTEGER log2 (const INTEGER& x)
{
	return MP_int_log(x.value);
}

//****************************************************************************************
// Absolute value: |x|
// returns the positive part of the object
// 1. argument: INTEGER
// return value: INTEGER
//****************************************************************************************

INTEGER abs (const INTEGER& x)
{
  MP_int_type zvalue;
  MP_int_init(zvalue);
  MP_int_abs(x.value,zvalue);
  return zvalue;
}


//****************************************************************************************
// boolean operators
// Last change: 28.01.2001
//****************************************************************************************

//****************************************************************************************
// Less than: x < y
//****************************************************************************************

bool operator < (const INTEGER& x, const INTEGER& y){
  return (MP_int_sign((y-x).value)==1);
}
bool operator <  (const INTEGER& x, const int      y){ return x<INTEGER(y) ;}
bool operator <  (const INTEGER& x, const long     y){ return x<INTEGER(y) ;}
bool operator <  (const int      x, const INTEGER& y){ return INTEGER(x)<y ;}
bool operator <  (const long     x, const INTEGER& y){ return INTEGER(x)<y ;}

bool operator <= (const INTEGER& x, const INTEGER& y){ return !(y<x) ;}
bool operator <= (const INTEGER& x, const int      y){ return !(y<x) ;}
bool operator <= (const INTEGER& x, const long     y){ return !(y<x) ;}
bool operator <= (const int      x, const INTEGER& y){ return !(y<x) ;}
bool operator <= (const long     x, const INTEGER& y){ return !(y<x) ;}

bool operator >  (const INTEGER& x, const INTEGER& y){ return y<x ;}
bool operator >  (const INTEGER& x, const int      y){ return y<x ;}
bool operator >  (const INTEGER& x, const long     y){ return y<x ;}
bool operator >  (const int      x, const INTEGER& y){ return y<x ;}
bool operator >  (const long     x, const INTEGER& y){ return y<x ;}

bool operator >= (const INTEGER& x, const INTEGER& y){ return !(x<y) ;}
bool operator >= (const INTEGER& x, const int      y){ return !(x<y) ;}
bool operator >= (const INTEGER& x, const long     y){ return !(x<y) ;}
bool operator >= (const int      x, const INTEGER& y){ return !(x<y) ;}
bool operator >= (const long     x, const INTEGER& y){ return !(x<y) ;}



bool operator == (const INTEGER& x, const INTEGER& y){
  return MP_int_sign((y-x).value)==0;
}

bool operator == (const INTEGER& x, const int  i){ return x==INTEGER(i);}
bool operator == (const INTEGER& x, const long i){ return x==INTEGER(i);}
bool operator == (const int      x, const INTEGER& y){ return (y==x) ;}
bool operator == (const long     x, const INTEGER& y){ return (y==x) ;}

bool operator != (const INTEGER& x, const INTEGER& y){ return !(x==y) ;} 
bool operator != (const INTEGER& x, const int      y){ return !(x==y) ;}
bool operator != (const INTEGER& x, const long     y){ return !(x==y) ;}
bool operator != (const int      x, const INTEGER& y){ return !(x==y) ;}
bool operator != (const long     x, const INTEGER& y){ return !(x==y) ;}

//****************************************************************************************
// output functions
//****************************************************************************************



//****************************************************************************************
// rwrite: writes INTEGER to STOUT
// output function for integer 
// 1. argument: INTEGER
//****************************************************************************************


void rwrite(const INTEGER& x)
{
  if (ACTUAL_STACK.inlimit) 
    {
      DEBUG1(0,"illegal output in continuous section!\n");
      return;
    }
  if (printfausgaben > printfaufrufe)  
    {
      printfaufrufe++;
      return;
    }
  printfausgaben++;
  printfaufrufe++;

  MP_int_printf(x.value);

}

//*****************************************************************************
// Function rwrite
// output function for integer 
// 1. argument: INTEGER
// 2. argument: long integer : width
//*****************************************************************************

void rwrite(const INTEGER& x, const long w)
{
  if (ACTUAL_STACK.inlimit) 
    {
      DEBUG1(0,"illegal output in continuous section!\n");
      return;
    }
  if (printfausgaben > printfaufrufe)  
    {
      printfaufrufe++;
      return;
    }
  printfausgaben++;
  printfaufrufe++;

  MP_int_writee(x.value,w);    
}


//*****************************************************************************
// Function swrite
// output function for integer 
// 1. argument: INTEGER
// 2. argument: long integer : width
// 	return:   char*
//*****************************************************************************

char* swrite(const INTEGER& x, const long w){
   	return MP_int_swritee(x.value,w);    
}

//*****************************************************************************
// Function swrite
// output function for integer 
// 1. argument: INTEGER
// 2. argument: long integer : width
//*****************************************************************************

char* swrite(const INTEGER& x){
	return MP_int_sprintf(x.value);    
}

